@@ -2,7 +2,6 @@ use std::cell::RefCell;
 use std::marker::PhantomData;
 use std::rc::Rc;
 
-use sycamore::generic_node::EventHandler;
 use sycamore::prelude::*;
 use wasm_bindgen::prelude::*;
 use wasm_bindgen::JsCast;
@@ -21,7 +20,7 @@ pub trait Integration {
 
     /// Get the click handler that is run when links are clicked.
 
-    fn click_handler(&self) -> Box<EventHandler>;
+    fn click_handler(&self) -> Box<dyn Fn(web_sys::Event)>;
 }
 
 thread_local! {
@@ -63,7 +62,7 @@ impl Integration for HistoryIntegration {
         closure.forget();
     }
 
-    fn click_handler(&self) -> Box<EventHandler> {
+    fn click_handler(&self) -> Box<dyn Fn(web_sys::Event)> {
         Box::new(|ev| {
             if let Some(a) = ev
                 .target()
@@ -83,28 +82,26 @@ impl Integration for HistoryIntegration {
                 }
 
                 let origin = a.origin();
-                let path = a.pathname();
+                let a_pathname = a.pathname();
                 let hash = a.hash();
 
                 let meta_keys_pressed = meta_keys_pressed(ev.unchecked_ref::<KeyboardEvent>());
-                if !meta_keys_pressed && Ok(origin) == location.origin() {
-                    if Ok(&path) != location.pathname().as_ref() {
+                if !meta_keys_pressed && location.origin() == Ok(origin) {
+                    if location.pathname().as_ref() != Ok(&a_pathname) {
                         // Same origin, different path.
                         ev.prevent_default();
                         PATHNAME.with(|pathname| {
                             let pathname = pathname.borrow().clone().unwrap_throw();
-                            let path = path.strip_prefix(&base_pathname()).unwrap_or(&path);
+                            let path = a_pathname
+                                .strip_prefix(&base_pathname())
+                                .unwrap_or(&a_pathname);
                             pathname.set(path.to_string());
 
                             // Update History API.
                             let window = web_sys::window().unwrap_throw();
                             let history = window.history().unwrap_throw();
                             history
-                                .push_state_with_url(
-                                    &JsValue::UNDEFINED,
-                                    "",
-                                    Some(pathname.get().as_str()),
-                                )
+                                .push_state_with_url(&JsValue::UNDEFINED, "", Some(&a_pathname))
                                 .unwrap_throw();
                             window.scroll_to_with_x_and_y(0.0, 0.0);
                         });
@@ -133,7 +130,11 @@ fn base_pathname() -> String {
             let base = base.unchecked_into::<HtmlBaseElement>().href();
 
             let url = web_sys::Url::new(&base).unwrap_throw();
-            url.pathname()
+            let mut pathname = url.pathname();
+            // Strip trailing `/` character from the pathname.
+            pathname.ends_with('/');
+            pathname.pop(); // Pop the `/` character.
+            pathname
         }
         _ => "".to_string(),
     }
@@ -143,7 +144,7 @@ fn base_pathname() -> String {
 pub struct RouterProps<R, F, G>
 where
     R: Route + 'static,
-    F: FnOnce(StateHandle<R>) -> Template<G>,
+    F: FnOnce(ReadSignal<R>) -> View<G>,
     G: GenericNode,
 {
     render: F,
@@ -154,7 +155,7 @@ where
 impl<R, F, G> RouterProps<R, F, G>
 where
     R: Route + 'static,
-    F: FnOnce(StateHandle<R>) -> Template<G>,
+    F: FnOnce(ReadSignal<R>) -> View<G>,
     G: GenericNode,
 {
     /// Create a new [`RouterProps`].
@@ -170,10 +171,10 @@ where
 /// The sycamore router component. This component expects to be used inside a browser environment.
 /// For server environments, see [`StaticRouter`].
 #[component(Router<G>)]
-pub fn router<R, F>(props: RouterProps<R, F, G>) -> Template<G>
+pub fn router<R, F>(props: RouterProps<R, F, G>) -> View<G>
 where
     R: Route + 'static,
-    F: FnOnce(StateHandle<R>) -> Template<G> + 'static,
+    F: FnOnce(ReadSignal<R>) -> View<G> + 'static,
 {
     let RouterProps {
         render,
@@ -241,7 +242,7 @@ where
 pub struct StaticRouterProps<R, F, G>
 where
     R: Route,
-    F: Fn(R) -> Template<G>,
+    F: Fn(R) -> View<G>,
     G: GenericNode,
 {
     render: F,
@@ -252,7 +253,7 @@ where
 impl<R, F, G> StaticRouterProps<R, F, G>
 where
     R: Route,
-    F: Fn(R) -> Template<G>,
+    F: Fn(R) -> View<G>,
     G: GenericNode,
 {
     /// Create a new [`StaticRouterProps`].
@@ -270,10 +271,10 @@ where
 /// This is useful for SSR where we want the HTML to be rendered instantly instead of waiting for
 /// the route preload to finish loading.
 #[component(StaticRouter<G>)]
-pub fn static_router<R, F>(props: StaticRouterProps<R, F, G>) -> Template<G>
+pub fn static_router<R, F>(props: StaticRouterProps<R, F, G>) -> View<G>
 where
     R: Route + 'static,
-    F: Fn(R) -> Template<G> + 'static,
+    F: Fn(R) -> View<G> + 'static,
 {
     let StaticRouterProps {
         render,
@@ -306,7 +307,36 @@ pub fn navigate(url: &str) {
         let window = web_sys::window().unwrap_throw();
         let history = window.history().unwrap_throw();
         history
-            .push_state_with_url(&JsValue::UNDEFINED, "", Some(pathname.get().as_str()))
+            .push_state_with_url(&JsValue::UNDEFINED, "", Some(url))
+            .unwrap_throw();
+        window.scroll_to_with_x_and_y(0.0, 0.0);
+    });
+}
+
+/// Navigates to the specified `url` without adding a new history entry. Instead, this replaces the
+/// current location with the new `url`. The url should have the same origin as the app.
+///
+/// This is useful for imperatively navigating to an url when using an anchor tag (`<a>`) is not
+/// possible/suitable (e.g. when submitting a form).
+///
+/// # Panics
+/// This function will `panic!()` if a [`Router`] has not yet been created.
+pub fn navigate_replace(url: &str) {
+    PATHNAME.with(|pathname| {
+        assert!(
+            pathname.borrow().is_some(),
+            "navigate_replace can only be used with a BrowserRouter"
+        );
+
+        let pathname = pathname.borrow().clone().unwrap_throw();
+        let path = url.strip_prefix(&base_pathname()).unwrap_or(url);
+        pathname.set(path.to_string());
+
+        // Update History API.
+        let window = web_sys::window().unwrap_throw();
+        let history = window.history().unwrap_throw();
+        history
+            .replace_state_with_url(&JsValue::UNDEFINED, "", Some(url))
             .unwrap_throw();
         window.scroll_to_with_x_and_y(0.0, 0.0);
     });
@@ -335,7 +365,7 @@ mod tests {
         }
 
         #[component(Comp<G>)]
-        fn comp(path: String) -> Template<G> {
+        fn comp(path: String) -> View<G> {
             let route = Routes::match_route(
                 &path
                     .split('/')
@@ -343,16 +373,16 @@ mod tests {
                     .collect::<Vec<_>>(),
             );
 
-            template! {
+            view! {
                 StaticRouter(StaticRouterProps::new(route, |route: Routes| {
                     match route {
-                        Routes::Home => template! {
+                        Routes::Home => view! {
                             "Home"
                         },
-                        Routes::About => template! {
+                        Routes::About => view! {
                             "About"
                         },
-                        Routes::NotFound => template! {
+                        Routes::NotFound => view! {
                             "Not Found"
                         }
                     }
@@ -361,17 +391,17 @@ mod tests {
         }
 
         assert_eq!(
-            sycamore::render_to_string(|| template! { Comp("/".to_string()) }),
+            sycamore::render_to_string(|| view! { Comp("/".to_string()) }),
             "Home"
         );
 
         assert_eq!(
-            sycamore::render_to_string(|| template! { Comp("/about".to_string()) }),
+            sycamore::render_to_string(|| view! { Comp("/about".to_string()) }),
             "About"
         );
 
         assert_eq!(
-            sycamore::render_to_string(|| template! { Comp("/404".to_string()) }),
+            sycamore::render_to_string(|| view! { Comp("/404".to_string()) }),
             "Not Found"
         );
     }
